/*
 * Copyright (c) 1991, 1994, 1995, 1996, 1998, 1999, 2001, 2004, 2005, 2006, 2009
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 1991, 1994, 1995, 1996, 1998, 1999, 2001, 2004, 2005, 2006, 2009\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: cf.c 37 2009-10-11 07:37:55Z leres $ (LBL)";
#endif

#include <sys/types.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "version.h"

static char *argv0;

extern char *optarg;
extern int optind, opterr;

int preserve = 0;
int strict = 0;
int utc = 0;
char *sfmt = "%b %e %H:%M:%S";
char *lfmt = "%b %e %H:%M:%S %Y";
char *fmt;

/* Forwards */
int doargs(int , char **);
void doone(FILE *, FILE *);
int main(int, char **);
int makeargv(const char *, char ***);
void usage(void) __attribute__((noreturn));

int
main(int argc, char **argv)
{
	register char *cp;
	register int n, status, didany;
	FILE *f;
	int targc, eargc;
	char **targv, **eargv;

	if ((cp = strrchr(argv[0], '/')) != NULL)
		argv0 = cp + 1;
	else
		argv0 = argv[0];

	/* Set default format */
	fmt = sfmt;

	/* Process args and setup optional file args */
	n = doargs(argc, argv);
	targc = argc - optind;
	targv = &argv[optind];

	/* If there were no command line args, try the environment */
	if (n == 0 && (cp = getenv("CFTIMEARGS")) != NULL) {
		eargc = makeargv(cp, &eargv);
		(void)doargs(eargc, eargv);
		if (eargc != optind)
			usage();
	}

	/* Process optional file args */
	status = 0;
	didany = 0;
	while (targc > 0) {
		f = fopen(*targv, "r");
		if (f) {
			doone(f, stdout);
			(void) fclose(f);
		} else {
			(void) fprintf(stderr, "%s: fopen: ", argv0);
			perror(*targv);
			status |= 1;
		}
		--targc;
		++targv;
		++didany;
	}

	/* Fall back to stdin if there were no file args */
	if (!didany)
		doone(stdin, stdout);
	exit(status);
}

/* Returns the number of flags processed */
int
doargs(int argc, char **argv)
{
	register int op, n;

	n = 0;
	opterr = 0;
	while ((op = getopt(argc, argv, "f:lpsu")) != EOF)
		switch (op) {

		case 'f':
			if (*optarg == '\0')
				fmt = sfmt;
			else
				fmt = optarg;
			++n;
			break;

		case 'l':
			fmt = lfmt;
			++n;
			break;

		case 'p':
			++preserve;
			++n;
			break;

		case 's':
			++strict;
			++n;
			break;

		case 'u':
			++utc;
			++n;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	return (n);
}

void
doone(FILE *fin, FILE *fout)
{
	time_t ts;
	register char *bp, *dotbp;
	register struct tm *tp;
	register int dot_count;
	char buf[1024];
	char tstr[128] = "";
	static time_t lastts = 0;

	while (fgets(buf, sizeof(buf), fin)) {
		bp = buf;
		dotbp = NULL;
		if (isdigit(*bp)) {
			ts = atol(bp);
			++bp;
			dot_count = 0;
			while (isdigit(*bp) || *bp == '.') {
				if (*bp == '.') {
					dotbp = bp;
					++dot_count;
				}
				++bp;
			}
			if (strict && (bp - buf < 9 || dot_count > 1 ||
			    (bp - buf > 10 && dot_count != 1))) {
				/* Doesn't look like a genuine timestamp -
				 * skip it.
				 */
				fputs(buf, fout);
				continue;
			}
			if (lastts != ts) {
				if (!utc)
					tp = localtime(&ts);
				else
					tp = gmtime(&ts);
				(void)strftime(tstr, sizeof(tstr), fmt, tp);
				lastts = ts;
			}
			fputs(tstr, fout);
			if (preserve && dotbp != NULL)
				bp = dotbp;
		}
		fputs(bp, fout);
	}
}

/* This version only handles ' */
int
makeargv(const char *str, char ***argvp)
{
	register int argc;
	register char *cp, ch;
	char **argv;
	static char *env;

	/* Make a copy we can hack up */
	env = strdup(str);
	if (env == NULL) {
		(void)fprintf(stderr, "%s: strdup failed\n", argv0);
		exit(1);
	}

	/* Count arguments */
	argc = 1;
	cp = env;
	while (*cp != '\0') {
		switch (*cp) {

		case ' ':
		case '\t':
			++argc;
			++cp;
			break;

		case '\\':
			++cp;
			if (*cp != '\0')
				++cp;
			break;

		case '\'':
		case '"':
			ch = *cp;
			++cp;
			while (*cp != '\0') {
				if (*cp == ch) {
					++cp;
					break;
				}
				if (*cp == '\\') {
					++cp;
					if (*cp == '\0')
						break;
				}
				++cp;
			}
			break;

		default:
			++cp;
			break;
		}
	}

	/* Room for argv[0] and a trailing NULL */
	argv = calloc(argc + 2, sizeof(*argv));
	if (env == NULL) {
		(void)fprintf(stderr, "%s: calloc failed\n", argv0);
		exit(1);
	}

	/* Build argv */
	*argvp = argv;
	argv[0] = argv0;
	argv[1] = env;
	argc = 2;
	cp = env;
	while (*cp != '\0') {
		switch (*cp) {

		case ' ':
		case '\t':
			*cp = '\0';
			++cp;
			argv[argc] = cp;
			++argc;
			break;

		case '\\':
			memmove(cp, cp + 1, strlen(cp));
			if (*cp != '\0')
				++cp;
			break;

		case '\'':
		case '"':
			ch = *cp;
			memmove(cp, cp + 1, strlen(cp));
			while (*cp != '\0') {
				if (*cp == ch) {
					memmove(cp, cp + 1, strlen(cp));
					break;
				}
				if (*cp == '\\') {
					memmove(cp, cp + 1, strlen(cp));
					if (*cp == '\0')
						break;
				}
				++cp;
			}
			break;

		default:
			++cp;
			break;
		}
	}
	return (argc);
}

void
usage()
{

	(void)fprintf(stderr, "%s version %s\n", argv0, version);
	(void)fprintf(stderr, "usage: %s [-f fmt] [-lpsu] [file ...]\n", argv0);
	exit(1);
}
